#include "arm9_write.h"

void WRHDECL arm9_write_byte_gen(unsigned int address,int word)
{
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFF;

	if (arm9_pageTable[pagen])
		if (arm9_pageLength[page]>offset)
			*(unsigned char*)(arm9_pageTable[pagen] + offset) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_dtcm(unsigned int address,int word)
{
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(unsigned char*)(DTCM + (address-dtcmBase)) = (unsigned char)word;
	else if (arm9_pageTable[address>>20])
		*(unsigned char*)(arm9_pageTable[address>>20] + (address&0xFFFFF)) = (unsigned char)word;
}

void WRHDECL arm9_write_byte_itcm(unsigned int address,int word)
{
	if ((address<0x8000))
		*(unsigned char*)(ITCM + address) = (unsigned char)word;
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(unsigned char*)(DTCM + (address-dtcmBase)) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_ptok(unsigned int address,int word) {
	*(arm9_pageTable[address>>20] + (address&0xFFFFF)) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_wram(unsigned int address,int word) {
	*(arm9_pageTable[address>>20] + (address&arm9_pageMask[address>>20])) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_sram(unsigned int address,int word) {
	*(arm9_pageTable[0xA] + (address&0xFFFF)) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_plt(unsigned int address,int word) {
	*(PLT_VRAM + (address&0xFFF)) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_oam(unsigned int address,int word) {
	*(OAM + (address&0x7FF)) = (unsigned char)word;
}


void WRHDECL arm9_write_byte_io(unsigned int address,int word) {
	unsigned int pagen=address>>20,offset=address&0x0FFFFF;

//	if (arm9_pageTable[pagen])
//		if (pageLength[4]>offset)
			*(unsigned char*)(arm9_pageTable[pagen] + offset) = (unsigned char)word;

		//wsprintf(buffer,"[%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	switch (offset)	{
		case IME:
			*(int*)(arm9_pageTable[0x40] + IME) &= 1;
			break;
		case IE:
			//wsprintf(buffer,"[%08X] = %02X, at %X\n",address,word&0xFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;

		case IF_:
			irqFlags &= 0xFF00|((word^0xFFFF)&0xFF);
			*(unsigned short int*)(arm9_pageTable[pagen] + offset) = irqFlags;
			break;
		/*case TM0CNT:
			timer_setup(0);
			break;
		case TM1CNT:
			timer_setup(1);
			break;
		case TM2CNT:
			timer_setup(2);
			break;
		case TM3CNT:
			timer_setup(3);
			break;*/

		case 0x180:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			/**(unsigned short*)(arm7_pageTable[pagen] + 0x180) &= 0xFFF0;
			*(unsigned char*)(arm7_pageTable[pagen] + 0x180) |= (word>>8)&0xF;*/
			*(arm9_pageTable[pagen] + 0x180) &= 0xF0;
			*(arm9_pageTable[pagen] + 0x180) |= ipcSync;
			//ipcSync7 = (word>>8)&0xF;
			wsprintf(buffer,"(9) Writing %08X at %X, (%X, %X)\n",
				word,arm9_regs[15],*(unsigned short int*)(arm9_pageTable[pagen] + 0x180),
				*(unsigned short int*)(arm7_pageTable[pagen] + 0x180));
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;
		case 0x184:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			//arm9FifoCnt &= ((word&0x4000)^0x4000);
			//*(unsigned short int*)(arm9_pageTable[0x40] + 0x184) = arm9FifoCnt;
			break;
		case 0x188:
			ipc_fifo_send(0,word&0xFF);
			break;

		case 0x240:
		case 0x241:
		case 0x242:
		case 0x243:
		case 0x244:
		case 0x245:
		case 0x246:
		case 0x248:
		case 0x249:
			remap_vram(offset - ((offset>0x246)?0x241:0x240),word&0xFF);
			break;

		case 0x247:
			IOREGS_ARM7[0x241] = word&0xFF;
			wsprintf(buffer,"WRAMCNT = %X (8bit) at %08X\n",(unsigned char)word,arm9_regs[15]);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			remap_wram(0);
			break;
	}
}


void WRHDECL arm9_write_byte_na(unsigned int address,int word)
{
}


void WRHDECL arm9_write_byte_vram(unsigned int address,int word)
{
	int chunk = (address>>14)&0x3FF;

	if (vramChunk[chunk]) {
		*(unsigned char*)(vramChunk[chunk] + (address&0x3FFF)) = (unsigned char)word;
	}
}

void WRHDECL arm9_write_byte_lcdc(unsigned int address,int word) {
	int chunk;
	address &= 0xFFFFF; if (address>0xA3FFF) address-=0xA3FFF; 
	chunk = ((address>>14)|0x200)&0x3FF;
	if (vramChunk[chunk]) {
		*(unsigned char*)(vramChunk[chunk] + (address&0x3FFF)) = (unsigned char)word;
	}
}


void WRHDECL arm9_write_byte_cf(unsigned int address,int word) {
	cflash_write(address,word);
}

//////////////////////////////////////////////////////////////////////////////////////////


void WRHDECL arm9_write_word_gen(unsigned int address,int word)
{
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFF;

	if (arm9_pageTable[pagen])
		if (arm9_pageLength[page]>offset)
			*(unsigned short int*)(arm9_pageTable[pagen] + offset) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_dtcm(unsigned int address,int word)
{
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(unsigned short int*)(DTCM + (address-dtcmBase)) = (unsigned short int)word;
	else if (arm9_pageTable[address>>20])
		*(unsigned short int*)(arm9_pageTable[address>>20] + (address&0xFFFFF)) = (unsigned short int)word;
}

void WRHDECL arm9_write_word_itcm(unsigned int address,int word)
{
	if ((address<0x8000))
		*(unsigned short int*)(ITCM + address) = (unsigned short int)word;
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(unsigned short int*)(DTCM + (address-dtcmBase)) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_ptok(unsigned int address,int word) {
	*(unsigned short int*)(arm9_pageTable[address>>20] + (address&0xFFFFF)) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_wram(unsigned int address,int word) {
	*(unsigned short int*)(arm9_pageTable[address>>20] + (address&arm9_pageMask[address>>20])) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_plt(unsigned int address,int word) {
	*(unsigned short int*)(PLT_VRAM + (address&0xFFF)) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_oam(unsigned int address,int word) {
	*(unsigned short int*)(OAM + (address&0x7FF)) = (unsigned short int)word;
}


void WRHDECL arm9_write_word_io(unsigned int address,int word)
{
	unsigned int pagen=address>>20,offset=address&0x0FFFFF;

//	if (arm9_pageTable[pagen])
//		if (pageLength[4]>offset)
			*(unsigned short int*)(arm9_pageTable[pagen] + offset) = (unsigned short int)word;

	/*if (address==0x40004A8) {
		wsprintf(buffer,"[%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

		//wsprintf(buffer,"[%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
		//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	switch (offset)	{
		case REG_DIVCNT:
			//wsprintf(buffer,"[%08X] = %04X, at %X\n",address,word,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			cp_div(word&3);
			break;
		case REG_NUMER_L:		
		case REG_NUMER_L+2:		
		case REG_DENOM_L:		
		case REG_DENOM_L+2:		
			cp_div(-1);
			break;

		case REG_SQRTCNT:
			cp_sqrt(word&3);
			break;
			
		case 0xBA:
			dma_setup(0,0,false);
			break;
		case 0xC6:
			dma_setup(0,1,false);
			break;
		case 0xD2:
			dma_setup(0,2,false);
			break;
		case 0xDE:
			dma_setup(0,3,false);
			break;

		case IME:
			*(int*)(arm9_pageTable[0x40] + IME) &= 1;
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;
		case IE:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;

		case IF_:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			irqFlags &= (word^0xFFFF);
			*(unsigned short int*)(arm9_pageTable[pagen] + offset) = irqFlags; //&0x1F7F;
			break;
		/*case TM0D:
			break;
		case TM1D:
			break;
		case TM2D:
			break;
		case TM3D:
			break;*/
		case TM0CNT:
			timer_setup(0,0);
			break;
		case TM1CNT:
			timer_setup(0,1);
			break;
		case TM2CNT:
			timer_setup(0,2);
			break;
		case TM3CNT:
			timer_setup(0,3);
			break;

		case 0x1C0:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;

		case 0x180:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			*(unsigned short int*)(arm7_pageTable[pagen] + 0x180) &= 0xFFF0;
			*(unsigned short int*)(arm7_pageTable[pagen] + 0x180) |= (word>>8)&0xF;
			*(arm9_pageTable[pagen] + 0x180) &= 0xF0;
			*(arm9_pageTable[pagen] + 0x180) |= ipcSync;
			ipcSync7 = (word>>8)&0xF;
			break;

		case 0x184:
			arm9FifoCnt &= ~0x8404;
			arm9FifoCnt |= (word&0x8404);
			arm9FifoCnt &= ((word&0x4000)^0x4000);
			*(unsigned short int*)(arm9_pageTable[0x40] + 0x184) = arm9FifoCnt;
			break;

		case 0x188:
			ipc_fifo_send(0,word&0xFFFF);
			break;

		case 0x23F:
		case 0x240:
		case 0x241:
		case 0x242:
		case 0x243:
		case 0x244:
		case 0x245:
		case 0x246:
		case 0x247:
		case 0x248:
		case 0x249:
			for (unsigned int i=offset; i<offset+2; i++)
			{
				if (i>0x249) break;
				if ((offset>=0x240)&&(offset != 0x247))
					remap_vram(i - ((offset>0x246)?0x241:0x240),word&0xFF);
				if (offset == 0x247) {
					wsprintf(buffer,"WRAMCNT = %X (16bit)\n",word&0xFF);
					WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					IOREGS_ARM7[0x241] = word&0xFF;
					remap_wram(0);
				}

				word >>= 8;
			}
			break;
	}
}


void WRHDECL arm9_write_word_na(unsigned int address,int word)
{
}


void WRHDECL arm9_write_word_vram(unsigned int address,int word)
{
	int chunk = (address>>14)&0x3FF;
	//dummy = address;

	//if (chunk==0x3)
	/*if ((address >= 0x600F800) && (address < 0x600FA00))
	{*/
	/*	wsprintf(buffer,"[%08X] = %04X, chunk=%d, [chunk]=%X, %X\n",address,word&0xFFFF,
			chunk,vramChunk[chunk],vramChunk[chunk] + (address&0x3FFF));
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
	/*}*/

	if (vramChunk[chunk])
	{
		/*if (((vramChunk[chunk]+(address&0x3FFF))>=VRAM_A+0x20000)&&
		    ((vramChunk[chunk]+(address&0x3FFF))<VRAM_A+0x20100))*/
		*(unsigned short int*)(vramChunk[chunk] + (address&0x3FFF)) = (unsigned short int)word;
	}
}

void WRHDECL arm9_write_word_lcdc(unsigned int address,int word)
{
	int chunk;
	address &= 0xFFFFF; if (address>0xA3FFF) address-=0xA4000; //if (address>0x9FFFF) address-=0x9FFFF;
	chunk = ((address>>14)|0x200)&0x3FF;
	//chunk = (address>>14)&0x3FF; 
	if (vramChunk[chunk])
		*(unsigned short int*)(vramChunk[chunk] + (address&0x3FFF)) = (unsigned short int)word;
}



void WRHDECL arm9_write_word_cf(unsigned int address,int word) {
	cflash_write(address,word);
}


/////////////////////////////////////////////////////////////////////////////////////////


void WRWDECL arm9_write_dword_gen(unsigned int address,int word)
{
	unsigned int page = address>>24,pagen=address>>20,offset=address&0x0FFFFC;
	//dummy = address;

	if (arm9_pageTable[pagen])
		if (arm9_pageLength[page]>offset)
			*(int*)(arm9_pageTable[pagen] + offset) = word;
}


void WRWDECL arm9_write_dword_dtcm(unsigned int address,int word) {
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(int*)(DTCM + ((address-dtcmBase)&0xFFFC)) = word;
	else if (arm9_pageTable[address>>20])
		*(int*)(arm9_pageTable[address>>20] + (address&0xFFFFC)) = word;
}

void WRHDECL arm9_write_dword_itcm(unsigned int address,int word) {
	if ((address<0x8000))
		*(int*)(ITCM + address) = word;
	if ((address>=dtcmBase) && (address<dtcmBase+0x4000))
		*(int*)(DTCM + (address-dtcmBase)) = word;
}

void WRWDECL arm9_write_dword_ptok(unsigned int address,int word) {
	*(int*)(arm9_pageTable[address>>20] + (address&0xFFFFC)) = word;
}


void WRWDECL arm9_write_dword_wram(unsigned int address,int word) {
	//*(int*)(IWRAM + ((address-0x78000)&0x7FFC)) = word;
	*(int*)(arm9_pageTable[address>>20] + (address&arm9_pageMask[address>>20]&0x7FFC)) = word;
}


void WRWDECL arm9_write_dword_plt(unsigned int address,int word) {
	*(int*)(PLT_VRAM + (address&0xFFC)) = word;
}


void WRWDECL arm9_write_dword_oam(unsigned int address,int word)
{
	*(int*)(OAM + (address&0x7FC)) = word;
}


void WRWDECL arm9_write_dword_io(unsigned int address,int word)
{
	unsigned int pagen=address>>20,offset=address&0x0FFFFC;

	*(int*)(arm9_pageTable[pagen] + offset) = word;

	/*if (address==0x40004A8) {
		wsprintf(buffer,"[%08X] = %08X, at %X, sp=%X\n",address,word,arm9_regs[15],arm9_regs[13]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	if ((offset>=0x320)&&(offset<=0x600))
	{
		gpu_set(GPU_3D_COMMAND,offset,word);
		return;
	}

	/*if ((offset >= 0x30)&&(offset<=0x40)) {
		wsprintf(buffer,"[%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
	}*/

	switch (offset)
	{
		case REG_DIVCNT:
			//wsprintf(buffer,"[%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			cp_div(word&3);
			break;
		case REG_NUMER_L:
		case REG_DENOM_L:
			cp_div(-1);
			break;
		case REG_SQRTPARM:
			cp_sqrt(-1);
			break;

		case TM0D:
		case TM0CNT:
			timer_setup(0,0);
			break;
		case TM1D:
		case TM1CNT:
			timer_setup(0,1);
			break;
		case TM2D:
		case TM2CNT:
			timer_setup(0,2);
			break;
		case TM3D:
		case TM3CNT:
			timer_setup(0,3);
			break;

		case IME:
			*(int*)(arm9_pageTable[0x40] + IME) &= 1;
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;

		case IF_-2:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			irqFlags &= (((word&0xFFFF)>>16)^0xFFFF);
			*(unsigned short int*)(arm9_pageTable[pagen] + IF_) = irqFlags;
			break;
		case IF_:
			//wsprintf(buffer,"(9) [%08X] = %04X, at %X\n",address,word&0xFFFF,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			irqFlags &= ((word&0xFFFF)^0xFFFF);
			*(unsigned short int*)(arm9_pageTable[pagen] + IF_) = irqFlags;
			break;

		case 0xB8:
			dma_setup(0,0,false);
			break;
		case 0xBA:
			dma_setup(0,0,false);
			break;
		case 0xC4:
			dma_setup(0,1,false);
			break;
		case 0xC6:
			dma_setup(0,1,false);
			break;
		case 0xD0:
			dma_setup(0,2,false);
			break;
		case 0xD2:
			dma_setup(0,2,false);
			break;
		case 0xDC:
			dma_setup(0,3,false);
			break;
		case 0xDE:
			dma_setup(0,3,false);
			break;

		/*case 0xD4:
			wsprintf(buffer,"(9) [%08X] = %08X, at %X, SP=%X, LR=%X\n",address,word,arm9_regs[15],arm9_regs[13],arm9_regs[14]);
			WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			break;*/

		case 0x180:
			//wsprintf(buffer,"(9) [%08X] = %08X, at %X\n",address,word,arm9_regs[15]);
			//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
			*(unsigned short int*)(arm7_pageTable[pagen] + 0x180) &= 0xFFF0;
			*(unsigned short int*)(arm7_pageTable[pagen] + 0x180) |= (word>>8)&0xF;
			*(arm9_pageTable[pagen] + 0x180) &= 0xF0;
			*(arm9_pageTable[pagen] + 0x180) |= ipcSync;
			ipcSync7 = (word>>8)&0xF;
			break;

		case 0x188:
			ipc_fifo_send(0,word);
			break;

		case 0x23D:
		case 0x23E:
		case 0x23F:
		case 0x240:
		case 0x241:
		case 0x242:
		case 0x243:
		case 0x244:
		case 0x245:
		case 0x246:
		case 0x247:
		case 0x248:
		case 0x249:
			for (unsigned int i=offset; i<offset+4; i++)	{
				if (i>0x249) break;
				if ((offset>=0x240)&&(offset != 0x247))
					remap_vram(i - ((offset>0x246)?0x241:0x240),word&0xFF);
				if (offset == 0x247) {
					wsprintf(buffer,"WRAMCNT = %X (32bit)\n",word&0xFF);
					WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);
					remap_wram(0);
					IOREGS_ARM7[0x241] = word&0xFF;
				}
				word >>= 8;
			}
			break;
	}
}


void WRWDECL arm9_write_dword_na(unsigned int address,int word)
{
}


void WRWDECL arm9_write_dword_vram(unsigned int address,int word)
{
	int chunk = (address>>14)&0x3FF;

	/*if (address<0x6005000)		
	{*/
	/*	wsprintf(buffer,"[%08X] = %08X, chunk=%d, [chunk]=%X, %X, %X\n",address,word,
			chunk,vramChunk[chunk],vramChunk[chunk] + (address&0x3FFF),vramBase[0]);
		WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);*/
	/*}*/


	if (vramChunk[chunk]) {
		*(int*)(vramChunk[chunk] + (address&0x3FFC)) = word;
	}
}

void WRWDECL arm9_write_dword_lcdc(unsigned int address,int word)
{
	int chunk;
	address &= 0xFFFFF; if (address>0xA3FFF) address-=0xA4000;
	chunk = ((address>>14)|0x200)&0x3FF;
	//chunk = (address>>14)&0x3FF;

	//wsprintf(buffer,"[%08X] = %08X\n",address,word);
	//WriteConsole(hout,buffer,lstrlen(buffer),(unsigned long*)&dummy,NULL);

	if (vramChunk[chunk]) {
		*(int*)(vramChunk[chunk] + (address&0x3FFC)) = word;
	}
}


void WRWDECL arm9_write_dword_cf(unsigned int address,int word) {
	cflash_write(address,word);
}

